/*
 * http.c
 *
 * Copyright (C) 1993, 1994, John Kilburg.
 *
 * You may distribute this code freely as long as it is used only
 * for academic, research, and internal business uses only, without
 * a fee.  You may distribute the binary and source code to third
 * parties provided that the copyright notice and this statement
 * appears on all copies and that no charge is associated with such
 * copies.
 *
 * If you make and distribute modified versions of the source or
 * binaries then you must inform me (John Kilburg) of the distribution.
 * You must also give me credit for the original in the source and
 * documentation but you must indicate that modifications were made.
 *
 * If you wish to make commercial use of the source or binaries you should    
 * contact me, to negotiate an appropriate license for such
 * commercial use.  Commercial use includes (1) integration of all or    
 * part of the source code into a product for sale or license by or on   
 * behalf of Licensee to third parties, or (2) distribution of the binary
 * code or source code to third parties that need it to utilize a          
 * commercial product sold or licensed by you or on your behalf.        
 * 
 * This software is provided "as is" without expressed or
 * implied warranty.  It is not intended to be used for any particular
 * purpose.  I shall not be liable for any damages suffered by users
 * of this software.  USE AT YOUR OWN RISK.
 *
 * Handles HTTP 'GET'.  Use this to grab a document from an HTTP server.
 * Handles HTTP/1.0.
 *
 */
#include <stdio.h>
#include <ctype.h>

#ifndef NOSTDHDRS
#include <stdlib.h>
#endif

#include "url.h"
#include "document.h"
#include "access.h"
#include "net.h"
#include "util.h"
#include "conf.h"

#if NeedFunctionPrototypes
static void standard_http(Document *);
extern int DisplayTransferStatus(int);
static void http_main(Document *, char *, int, char *, char *, char *);
#else
static void standard_http();
extern int DisplayTransferStatus();
static void http_main();
#endif

/*
 * http_proxy
 *
 * Glue another interface onto http.  Of course, this will have to change
 * someday because the standard will keep changing and changing and
 * changing...
 */
void
http_proxy(d, hostname, portno, filename, ext, reload)
Document *d;
char *hostname;
int portno;
char *filename;
char *ext;
int reload;
{
  char *extra;

  if (reload)
  {
    extra = "Pragma: no-cache\r\n";
  }
  else
  {
    extra = "";
  }

  http_main(d, hostname, portno, filename, ext, extra);

  return;
}

/*
 * http
 *
 * The reasonable interface.
 */
void
http(d, hostname, portno, filename, ext)
Document *d;
char *hostname;
int portno;
char *filename;
char *ext;
{
  http_main(d, hostname, portno, filename, ext, "");

  return;
}

/*
 * http_main
 */
static void
http_main(d, hostname, portno, filename, ext, extra)
Document *d;
char *hostname;
int portno;
char *filename;
char *ext;
char *extra;
{
  char *query;
  char buffer[BUFSIZ];
  int blen;
  int s;
  int count = 0;
  char *t = NULL;
  int tlen = 0;
  int btlen = 0;
  static char *format_get = "GET %s HTTP/1.0\r\nUser-Agent: Chimera-HTTP/1.0\r\nAccept: */*\r\n%s\r\n";
  static char *format_get2 = "GET %s?%s HTTP/1.0\r\nUser-Agent: Chimera-HTTP/1.0\r\nAccept: */*\r\n%s\r\n";
  static char *format_post = "POST %s HTTP/1.0\r\nUser-Agent: Chimera-HTTP/1.0\r\nContent-length: %d\r\nContent-type: application/x-www-form-urlencoded\r\n%s\r\n%s\r\n";

  if (portno == -1)
  {
    portno = DEFAULT_HTTP_PORT;
  }

  d->type = DocUnknown;
  d->text = NULL;

  /*
   * Start chitchatting with the HTTP server.
   */
  s = net_open(hostname, portno);
  if (s < 0)
  {
    return;
  }

  /*
   * Send request
   */
  if (ext[0] == '?')
  {
    if (ext[1] == 'G') /* GET ! */
    {
      query = alloc_string_mem(strlen(filename) + strlen(format_get2) +
			       strlen(ext) + strlen(extra) + 1);
      sprintf (query, format_get2, filename, ext + 2, extra);
    }
    else if (ext[1] == 'P') /* POST ! */
    {
      query = alloc_string_mem(strlen(filename) + strlen(format_post) +
			       strlen(ext + 1) + strlen(extra) + 1);
      sprintf (query, format_post, filename, strlen(ext + 2), extra, ext + 2);
    }
    else
    {
      query = alloc_string_mem(strlen(filename) + strlen(format_get2) +
			       strlen(ext) + strlen(extra) + 1);
      sprintf (query, format_get2, filename, ext + 1, extra);
    }
  }
  else /* regular GET */
  {
    query = alloc_string_mem(strlen(filename) + strlen(format_get) +
			     strlen(ext) + 1);
    sprintf (query, format_get, filename, extra);
  }

  if (net_write(s, query, strlen(query)) < 0)
  {
    net_close(s);
    free(query);

    return;
  }

  free(query);

  /*
   * Read info from the HTTP host
   */
  while ((blen = net_read(s, buffer, sizeof(buffer))) > 0)
  {
    tlen += blen;
    if (t)
    {
      t = (char *)realloc_mem(t, tlen + 1);
    }
    else
    {
      t = (char *)alloc_mem(tlen + 1);
    }
    
    memcpy(t + btlen, buffer, blen);
    btlen = tlen;

    if (count++ % 10 == 0)
    {
      if (DisplayTransferStatus(tlen) == 1)
      {
	free(t);
	
	d->text = alloc_string("<h1>Info</h1>Transfer cancelled.");
	d->type = DocInternal;
	d->len = strlen(d->text);

	net_close(s);
	
	return;
      }
    }
  }

  net_close(s);

  if (tlen == 0)
  {
    return;
  }

  /*
   * Terminate the data, fill in the Document.
   */
  t[tlen] = '\0';
  d->text = t;
  d->len = tlen;

  /*
   * Check for the proper header information.  Notice the evil 5.
   * Everyone should be using HTTP/1.0 by now.
   */
  if (mystrncmp("HTTP/", d->text, 5) == 0)
  {
    standard_http(d);
  } 
  else
  {
    if (filename[strlen(filename) - 1] == '/')
    {
      d->content = alloc_string("text/html");
    }
    else
    {
      char *cp;

      for (cp = filename; *cp; cp++)
      {
	if (*cp == '?')
	{
	  d->content = alloc_string("text/html");
	  break;
	}
      }
    }
  }

  return;
}

/*
 * standard_http
 *
 * This function sorts out the standard HTTP header (which looks like
 * a MIME header and sorts out the error code and all that jazz.
 * This is a whole bunch of hard-coded MIME string fields and stuff
 * in here.
 */
static void
standard_http(d)
Document *d;
{
  char buffer[256];
  char junk[256];
  char field[256];
  char data[256];
  char *t, *s;
  int error_code;
  int i, j;

  /*
   * The first line should be a status line:
   *
   * HTTP/x.x error_code swell message with spaces
   *
   * In fact, we shouldn't even get here if the line didn't start "HTTP"...
   */
  t = get_line(d->text, buffer, sizeof(buffer));
  if (t == NULL)
  {
    /*
     * Unlikely to happen but you never know...
     */
    return;
  }

  sscanf (buffer, "%s %d %s", junk, &error_code, junk);

  /*
   * Read the MIME fields and figure out what is going on.
   */
  while ((t = get_line(t, buffer, sizeof(buffer))) != NULL)
  {
    /*
     * Check for end of MIME header (that's a blank line to you and me).
     */
    if (buffer[0] == '\0')
    {
      break;
    }

    /*
     * Oversimplified MIME field parser.  But heck we are only looking
     * for MIME fields with simple data.
     */
    sscanf(buffer, "%[^:] %[:] %[^\n]", field, junk, data);

    /*
     * Pick out the MIME fields that we understand and do stuff with
     * the data.  Only understands Content-type and Location.  It may
     * not handle them correctly since I didn't look at the RFC.
     */
    if (mystrncmp("Content-type", field, 12) == 0)
    {
      /*
       * Ho boy!  This is good.  Now we know how to deal with the document.
       * If the string allocate fails then it will look at the filename
       * so it isn't a complete disaster.  Of course if this puny
       * allocate fails then something else is bound to fail.
       */
      d->content = alloc_string(data);
    }
    else if (mystrncmp("Content-transfer-encoding", field, 8) == 0)
    {
      d->tencoding = alloc_string(data);
    }
    else if (mystrncmp("Content-encoding", field, 8) == 0)
    {
      d->encoding = alloc_string(data);
    }
    else if (mystrncmp("Location", field, 8) == 0)
    {
      /*
       * Handle 3xx guys.  Try to deal with looping.  Of course a document
       * could go to
       * another document which loops back to the current document.
       * I don't care to be that clever.  I figure this will fool the
       * BSDI test case.
       */
      if (error_code >= 300 && error_code <= 399)
      {
	s = MakeURL(d->up, 0);
	if (mystrcmp(s, data))
	{
	  Document *t;
	  URLParts *up;

	  up = MakeURLParts(data, NULL);
	  t = LoadDocument(up);
	  if (t != NULL)
	  {
	    TransferDocument(d, t);
	  }
	  DestroyURLParts(up);

	  free(s);
	  return;
	}
	free(s);
      }
    }
  }

  /*
   * It is possible that the info is screwed up.  Just return and let
   * the user see the server's screwup (it might be the client's fault).
   */
  if (t == NULL)
  {
    return;
  }

  /*
   * Since the amount of actual data is smaller we can reuse the
   * space.  Have to do a copy anyways.
   */
  for (j = 0, i = t - d->text; i < d->len; i++, j++)
  {
    d->text[j] = d->text[i]; 
  }
  d->text[j] = '\0';
  d->len = j;

  if (error_code >= 400)
  {
    d->content = alloc_string("text/html");
  }

  return;
}



